/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.
  Copyright (c) 2005 Mellanox Technologies Ltd.  All rights reserved.

  $Id: config.c 1897 2005-02-24 15:43:29Z mst $
*/

/*
 * system includes
 */
#if HAVE_CONFIG_H
#  include <config.h>
#endif /* HAVE_CONFIG_H */

#include <unistd.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <arpa/inet.h>
/*
 * SDP specific includes
 */
#include "libsdp.h"

/* --------------------------------------------------------------------- */
/* library static and global variables                                   */
/* --------------------------------------------------------------------- */
struct match_group *__sdp_match_groups_head = NULL;
struct match_group *__sdp_match_groups_tail = NULL;
struct match_group *__sdp_match_both_groups_head = NULL;
struct match_group *__sdp_match_both_groups_tail = NULL;

int __sdp_config_empty(
                       void
                       )
{
  return (__sdp_match_groups_head == NULL &&  __sdp_match_both_groups_head == NULL) ;
}

static char *next_token(
                        char *buf
                        )
{
  char *p;

  p = strpbrk(buf, " \t");
  if (!p) {
    p = strchr(buf, 0);
  } /* if */

  if (*p) {
    *p++ = 0;

    while (*p && strchr(" \t", *p)) {
      p++;
    } /* while */
  } /* if */

  return p;
}

static void group_free(
                       struct match_group *group, int match_both
                       )
{
  struct match_expression *expression, *nexpression;

  for (expression = group->head; expression; expression = nexpression) {
    nexpression = expression->next;

    free(expression);
  }
  if (match_both) {

    if (!group->prev) {
      __sdp_match_groups_head = group->next;
    } else {
      group->prev->next = group->next;
    }
    
    if (!group->next) {
      __sdp_match_groups_tail = group->prev;
    } else {
      group->next->prev = group->prev;
    }
  } else {
    if (!group->prev) {
      __sdp_match_both_groups_head = group->next;
    } else {
      group->prev->next = group->next;
    }
    
    if (!group->next) {
      __sdp_match_both_groups_tail = group->prev;
    } else {
      group->next->prev = group->prev;
    }
  }
  free(group);
}

static char *parse_ip_port(
                           struct match_expression *expression,
                           char *data
                           )
{
  char *buf, *port, *prefix, *pv;

  buf = strdup(data);
  if (!buf) {
    __sdp_log(9, "Unable to allocate %d bytes to parse data\n",
              strlen(data) + 1);
    return NULL;
  }

  pv = next_token(buf);

  /* No colon means match all ports */
  port = strchr(buf, ':');
  if (!port) {
    expression->data.ip_port.sport = 0x0000;
    expression->data.ip_port.eport = 0xFFFF;
  } else {
    *port++ = 0;

    if (strcmp(port, "*") == 0) {
      expression->data.ip_port.sport = 0x0000;
      expression->data.ip_port.eport = 0xFFFF;
    } else {
      char *dash;

      dash = strchr(port, '-');
      if (!dash) {
        unsigned long num;

        num = strtoul(port, NULL, 0);
        if (num == ULONG_MAX) {
          /* FIXME: Convert names from /etc/services to numbers here */
          __sdp_log(9, "Unknown port '%s'\n", port);
          return NULL;
        } /* if */

        expression->data.ip_port.sport = expression->data.ip_port.eport = num;
      } else {
        unsigned long svalue, evalue;

        *dash++ = 0;

        svalue = strtoul(port, NULL, 0);
        if (svalue == ULONG_MAX) {
          /* FIXME: Make a call to getservbyname() here? */
          __sdp_log(9, "Unable to parse start port '%s'\n", port);
          return NULL;
        } /* if */
        if (svalue > 0xFFFF) {
          __sdp_log(9, "Start port '%s' too large\n", port);
          return NULL;
        } /* if */

        evalue = strtoul(dash, NULL, 0);
        if (evalue == ULONG_MAX) {
          /* FIXME: Make a call to getservbyname() here? */
          __sdp_log(9, "Unable to parse end port '%s'\n", dash);
          return NULL;
        } /* if */
        if (evalue > 0xFFFF) {
          __sdp_log(9, "End port '%s' too large\n", dash);
          return NULL;
        } /* if */

        if (svalue > evalue) {
          __sdp_log(9, "Start port '%s' greater than end port '%s'\n", port, dash);
          return NULL;
        } /* if */

        expression->data.ip_port.sport = svalue;
        expression->data.ip_port.eport = evalue;
      }
    }
  }

  if (strcmp(buf, "*") == 0) {
    expression->data.ip_port.ip.s_addr = 0;
    expression->data.ip_port.prefixlen = 0;
  } else {
    prefix = strchr(buf, '/');
    if (prefix) {
      unsigned long length;

      *prefix++ = 0;

      length = strtoul(prefix, NULL, 0);
      if (length == ULONG_MAX || length > 32) {
        __sdp_log(9, "Unable to parse prefix length '%s'\n", prefix);
        return NULL;
      }

      expression->data.ip_port.prefixlen = length;
    } else {
      expression->data.ip_port.prefixlen = 32;
    } /* if */

    if (!inet_aton(buf, &expression->data.ip_port.ip)) {
      /* FIXME: Make a call to gethostbyname() here? */
      __sdp_log(9, "Unable to parse IP '%s'\n", buf);
      return NULL;
    }
  }

  expression->data.ip_port.ip.s_addr = expression->data.ip_port.ip.s_addr &
	htonl(SDP_NETMASK(expression->data.ip_port.prefixlen));

  return pv;
}

static char *parse_program_name(
                                struct match_expression *expression,
                                char *data
                                )
{
  char *pv;

  pv = next_token(data);

  if (!strlen(data)) {
    __sdp_log(9, "program_name is empty\n");
    return NULL;
  }

  expression->data.program_name = strdup(data);
  if (!expression->data.program_name) {
    __sdp_log(9, "Unable to allocate %d bytes memory for program_name\n",
            strlen(data) + 1);
    return NULL;
  }

  return pv;
}

static char *parse_match_expression(
                                    struct match_group *group,
                                    char *start
                                    )
{
  struct match_expression *expression;
  char *pn;

  pn = next_token(start);

  expression = malloc(sizeof *expression);
  if (!expression) {
    __sdp_log(9, "Unable to allocate %d bytes for expression\n",
            sizeof *expression);
    return NULL;
  } /* if */

  memset(expression, 0, sizeof *expression);

  expression->prev = group->tail;
  if (!group->head) {
    group->head = expression;
  } else {
    group->tail->next = expression;
  } /* if */
  group->tail = expression;

  if (strcasecmp(start, "listen") == 0) {
    expression->type = MATCH_LISTEN_IP_PORT;

    start = parse_ip_port(expression, pn);
    if (!start) {
      __sdp_log(9, "Unable to parse listen expression '%s'\n", pn);
      return NULL;
    }
  } else if (strcasecmp(start, "destination") == 0) {
    expression->type = MATCH_DESTINATION_IP_PORT;

    start = parse_ip_port(expression, pn);
    if (!start) {
      __sdp_log(9, "Unable to parse destination expression '%s'\n", pn);
      return NULL;
    }
  } else if (strcasecmp(start, "program") == 0) {
    expression->type = MATCH_PROGRAM_NAME;

    start = parse_program_name(expression, pn);
    if (!start) {
      __sdp_log(9, "Unable to parse program expression '%s'\n", pn);
      return NULL;
    }
  } else {
    __sdp_log(9, "Unknown token '%s'\n", start);
    return NULL;
  } /* if */

  return start;
}

static int parse_match(
                       char *data
                       )
{
  struct match_group *group;
  char *start = data;

  group = malloc(sizeof *group);
  if (!group) {
    __sdp_log(9, "Unable to allocate %d bytes for match_group\n",
            sizeof *group);
    return -1;
  } /* if */

  memset(group, 0, sizeof *group);

  group->prev = __sdp_match_groups_tail;
  if (!__sdp_match_groups_head) {
    __sdp_match_groups_head = group;
  } else {
    __sdp_match_groups_tail->next = group;
  } /* if */
  __sdp_match_groups_tail = group;

  /* Loop through parsing each expression */
  while (*start) {
    while (*start && strchr(" \t", *start)) {
      start++;
    }

    start = parse_match_expression(group, start);
    if (!start) {
      goto error;
    }
  } /* while */

  return 0;

error:
  __sdp_log(9, "Parsing match group failed, ignoring entire group\n");

  group_free(group, 0);

  return 1;
}



static int parse_match_both(
                       char *data
                       )
{
  struct match_group *group;
  char *start = data;

  group = malloc(sizeof *group);
  if (!group) {
    __sdp_log(9, "Unable to allocate %d bytes for match_group\n",
            sizeof *group);
    return -1;
  } /* if */

  memset(group, 0, sizeof *group);

  group->prev = __sdp_match_groups_tail;
  if (!__sdp_match_both_groups_head) {
    __sdp_match_both_groups_head = group;
  } else {
    __sdp_match_both_groups_tail->next = group;
  } /* if */
  __sdp_match_both_groups_tail = group;

  /* Loop through parsing each expression */
  while (*start) {
    while (*start && strchr(" \t", *start)) {
      start++;
    }

    start = parse_match_expression(group, start);
    if (!start) {
      goto error;
    }
  } /* while */

  return 0;

error:
  __sdp_log(9, "Parsing match group failed, ignoring entire group\n");

  group_free(group, 1);

  return 1;
}



static char *parse_log_expression(
                                  char *start
                                  )
{
  char *pn;

  pn = next_token(start);

  if (strcasecmp(start, "min-level") == 0) {
    start = next_token(pn);
    if (!start) {
      __sdp_log(9, "Unable to parse min-level '%s'\n", pn);
      return NULL;
    }

    __sdp_log_set_min_level(atoi(pn));
  } else if (strcasecmp(start, "destination") == 0) {
    start = next_token(pn);
    if (!start) {
      __sdp_log(9, "Unable to parse destination '%s'\n", pn);
      return NULL;
    }

    if (strcasecmp(pn, "stderr") == 0) {
      __sdp_log_set_log_stderr();
    } else if (strcasecmp(pn, "syslog") == 0) {
      __sdp_log_set_log_syslog();
    } else if (strcasecmp(pn, "file") == 0) {
      char *fpn;

      fpn = start;
      start = next_token(fpn);
      if (!start) {
        __sdp_log(9, "Unable to parse filename for log style 'file'\n");
        return NULL;
      }

      __sdp_log_set_log_file(fpn);
    } else {
      __sdp_log(9, "Unable to parse destination '%s'\n", pn);
      return NULL;
    }
  } else {
    __sdp_log(9, "Unknown token '%s'\n", start);
    return NULL;
  } /* if */

  return start;
}

static int parse_log(
                     char *data
                     )
{
  char *start = data;

  /* Loop through parsing each expression */
  while (*start) {
    while (*start && strchr(" \t", *start)) {
      start++;
    }

    start = parse_log_expression(start);
    if (!start) {
      return 1;
    }
  } /* while */

  return 0;
}

int __sdp_read_config(
                      char *config_file
                      )
{
  FILE *f;

  f = fopen(config_file, "r");
  if (!f) {
    return -1;
  } /* if */

  while (!feof(f)) {
    char buffer[1024];
    char *start, *p;

    buffer[0] = 0;
    if (!fgets(buffer, sizeof(buffer) - 1, f)) {
      break;
    } /* if */

    if (strlen(buffer) > 0 && buffer[strlen(buffer) - 1] == '\n') {
      buffer[strlen(buffer) - 1] = 0;
    } /* if */

    /* Search for a # first to see if there is a comment on this line */
    p = strchr(buffer, '#');
    if (p) {
      *p = 0;
    } /* if */

    start = buffer;

    /* Skip over all beginning whitespace */
    while (*start && strchr(" \t", *start)) {
      start++;
    } /* while */

    /* Skip any blank lines */
    if (!strlen(start)) {
      continue;
    } /* if */

    /* Parse out the first token */
    p = next_token(start);
    if (strcasecmp(start, "match_both") == 0) {
      parse_match_both(p);
    } else if (strcasecmp(start, "match") == 0 ||  strcasecmp(start, "match_sdp") ==0 ) {
      parse_match(p);
    } else if (strcasecmp(start, "log") == 0) {
      parse_log(p);
    } else {
      __sdp_log(9, "Unknown command '%s'\n", start);
    }
  }

  fclose(f);

  return 0;
}
