/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Topspin Communications.  All rights reserved.
  Copyright (c) 2005 Mellanox Technologies Ltd.  All rights reserved.

  $Id: port.c 2890 2005-07-21 11:57:18Z mst $
*/

/*
 * system includes
 */
#if HAVE_CONFIG_H
#  include <config.h>
#endif /* HAVE_CONFIG_H */

#include <unistd.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#define __USE_GNU
#define _GNU_SOURCE	/* define RTLD_NEXT */
#include <dlfcn.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <stdarg.h>
/*
 * SDP specific includes
 */
#include "libsdp.h"

#define MAPPED_SOCKET_MAX	1024
#define MAX_BIND_TRIES		65536

void __attribute__ ((constructor)) __sdp_init(void);
void __attribute__ ((destructor)) __sdp_fini(void);



/* --------------------------------------------------------------------- */
/* library type definitions.                                             */
/* --------------------------------------------------------------------- */

typedef int (*ioctl_func_t)(int fd, int request, void *arg0, void *arg1, void *arg2, void *arg3, void *arg4, void *arg5, void *arg6, void *arg7);
typedef int (*fcntl_func_t)(int fd, int cmd, ...);
typedef int (*socket_func_t)(int domain, int type, int protocol);
typedef int (*setsockopt_func_t)(int s, int level, int optname, const void *optval, socklen_t optlen);
typedef int (*connect_func_t)(int sockfd, const struct sockaddr *serv_addr, socklen_t addrlen);
typedef int (*listen_func_t)(int s, int backlog);
typedef int (*bind_func_t)(int sockfd, const struct sockaddr *my_addr, socklen_t addrlen);
typedef int (*close_func_t)(int fd);
typedef int (*dup_func_t)(int fd);
typedef int (*getsockname_func_t)(int fd, struct sockaddr *name, socklen_t *namelen);
typedef int (*getpeername_func_t)(int fd, struct sockaddr *name, socklen_t *namelen);
typedef int (*accept_func_t)(int fd, struct sockaddr *addr, socklen_t *addrlen); 
typedef int (*select_func_t)(int n,  fd_set  *readfds,  fd_set  *writefds,
       fd_set *exceptfds, struct timeval *timeout);
  


struct socket_lib_funcs {
  ioctl_func_t       ioctl;
  fcntl_func_t       fcntl;
  socket_func_t      socket;
  setsockopt_func_t  setsockopt;
  connect_func_t     connect;
  listen_func_t      listen;
  bind_func_t        bind;
  close_func_t       close;
  dup_func_t         dup;
  getpeername_func_t getpeername;
  getsockname_func_t getsockname;
  accept_func_t      accept;
  select_func_t      select;
}; /* socket_lib_funcs */

#define STATE_LISTEN_CALLED	1
#define STATE_BIND_CALLED	2

static int simple_sdp_library;

/* --------------------------------------------------------------------- */
/* library static and global variables                                   */
/* --------------------------------------------------------------------- */
extern char *program_invocation_name, *program_invocation_short_name;

#ifdef RTLD_NEXT
static void *__libc_dl_handle = RTLD_NEXT;
#else
static void *__libc_dl_handle;
#endif
/*
 * SDP socket info.
 */
struct sdp_socket_info {
  int fd;
  int state;
  int refcount;
  struct sockaddr_in *addr;
  int turn;
}; /* sdp_socket_info */

/* Indexed by the AF_INET fd */
static struct sdp_socket_info *__sdp_struct_list[MAPPED_SOCKET_MAX];


static struct socket_lib_funcs _socket_funcs = {
  .socket = NULL,
  /* Automatically sets all other elements to NULL */
}; /* _socket_funcs */


/* --------------------------------------------------------------------- */
/* Prototypes                                                            */
/* --------------------------------------------------------------------- */
void __sdp_init(
                void
                );

/* --------------------------------------------------------------------- */
/*                                                                       */
/* local static functions.                                               */
/*                                                                       */
/* --------------------------------------------------------------------- */

/* ========================================================================= */
/*..__sdp_id_map -- given a AF_INET fd return AF_INET_SDP fd */
static int __sdp_id_map(
			int fd
			)
{
  struct sdp_socket_info *sdp_sock_info;

  if (!(0 > fd) &&
      MAPPED_SOCKET_MAX > fd) {

    sdp_sock_info = __sdp_struct_list[fd];

    return ((NULL == sdp_sock_info) ? -1 : sdp_sock_info->fd);
  } /* if */
  else {

    return -1;
  } /* else */
} /* __sdp_id_map */

/* ========================================================================= */
/*..__sdp_sock_get -- rgiven a AF_INET fd return an AF_INET_SDP sock struct */
static struct sdp_socket_info * __sdp_sock_get(
					       int fd
					       )
{
  if (!(0 > fd) &&
      MAPPED_SOCKET_MAX > fd) {

    return __sdp_struct_list[fd];
  } /* if */
  else {

    return NULL;
  } /* else */
} /* __sdp_sock_get */

/* ========================================================================= */
/*..__sdp_err_cleanup -- an error occured on an SDP socket, cleanup */
static int __sdp_err_cleanup(
			     int fd
			     )
{
  struct sdp_socket_info *sdp_sock_info;
  int counter;
  int current;

  sdp_sock_info = __sdp_sock_get(fd);
  if (NULL == sdp_sock_info) {
    return 0;
  } /* if */
  /*
   * loop through sockets to make sure all references have been cleaned.
   * start at the first reference, since a single reference is the most
   * common case.
   */
  for (counter = 0, current = fd; 
       counter < MAPPED_SOCKET_MAX && 0 < sdp_sock_info->refcount;
       counter++) {
 
    if (sdp_sock_info == __sdp_struct_list[current]) {

      sdp_sock_info->refcount   -= 1;
      __sdp_struct_list[current] = NULL;
    } /* if */

    current += 1;
    current  = (current < MAPPED_SOCKET_MAX) ? current : 0;
  } /* for */

  _socket_funcs.close(sdp_sock_info->fd);

  if (NULL != sdp_sock_info->addr) {

    free(sdp_sock_info->addr);
  } /* if */

  free(sdp_sock_info);

  return 0;
} /* __sdp_err_cleanup */

/* ========================================================================= */
/*..decide_family -- determine the which socket should be used */
static int decide_family(
			 int fd, 
			 int dir, 
			 const struct sockaddr_in *sin
			 )
{
  struct sdp_socket_info *sdp_sock_info;
  int counter;
  int current;
  int match;

  sdp_sock_info = __sdp_sock_get(fd);
  if (NULL == sdp_sock_info) {
    return 0;
  } /* if */

  /* Now we need to check which socket we'll end up using */

  if (dir) {
    match = __sdp_match_connect(sin);
  } else {
    match = __sdp_match_listen(sin,0);
  } /* if */

  for (counter = 0, current = fd; 
       counter < MAPPED_SOCKET_MAX && 0 < sdp_sock_info->refcount;
       counter++) {

    if (sdp_sock_info == __sdp_struct_list[current]) {
	
      __sdp_log(4, "MATCH:      %s for fd <%d:%d>, using %s socket\n", 
		(match ? "positive" : "negative"), current, sdp_sock_info->fd,
		(match ? "SDP" : "TCP"));

      sdp_sock_info->refcount   -= 1;
      __sdp_struct_list[current] = NULL;

      if (match) {

	_socket_funcs.close(current); //TODO :fix
	dup2(sdp_sock_info->fd, current);
      } /* if */
    } /* if */

    current += 1;
    current  = (current < MAPPED_SOCKET_MAX) ? current : 0;
  } /* for */

  _socket_funcs.close(sdp_sock_info->fd);

  if (NULL != sdp_sock_info->addr) {

    free(sdp_sock_info->addr);
  } /* if */

  free(sdp_sock_info);
  if (match) {
    return 1;
  };
  return 0;
} /* decide_family */

/* --------------------------------------------------------------------- */
/*                                                                       */
/* Socket library function overrides.                                    */
/*                                                                       */
/* --------------------------------------------------------------------- */

/* ========================================================================= */
/*..ioctl -- replacement ioctl call. */
int ioctl(
          int fd,
          int request,
          void *arg0,
          void *arg1,
          void *arg2,
          void *arg3,
          void *arg4,
          void *arg5,
          void *arg6,
          void *arg7
          )
{
  struct sdp_socket_info *sdp_sock_info;
  int sret;
  int ret = 0;


  if (NULL == _socket_funcs.ioctl) {
    return -1;
  } /* if */

  sdp_sock_info = __sdp_sock_get(fd);

  __sdp_log(1, "IOCTL:      <%s:%d:%d> request <%d>\n",
            program_invocation_short_name, fd, __sdp_id_map(fd), request);

  ret = _socket_funcs.ioctl(fd, request,
			    arg0, arg1, arg2, arg3, arg4, arg5, arg6, arg7);
  if (!(0 > ret) &&
      NULL != sdp_sock_info) {

    sret = _socket_funcs.ioctl(sdp_sock_info->fd, request,
			       arg0, arg1, arg2, arg3, arg4, arg5, arg6, arg7);
    if (sret < 0) {
      
      __sdp_log(9, "Error %d calling ioctl for SDP socket, closing\n", errno);
      __sdp_err_cleanup(fd);
    } /* if */
  } /* if */

  __sdp_log(1, "IOCTL:      <%s:%d:%d> result <%d>\n",
            program_invocation_short_name, fd, __sdp_id_map(fd), ret);

  return ret;
} /* ioctl */

/* ========================================================================= */
/*..fcntl -- replacement fcntl call. */
int fcntl(
          int fd,
          int cmd,
          ...
          )
{
  struct sdp_socket_info *sdp_sock_info;
  int sret;
  int ret = 0;

  void *arg;
  va_list ap;
  va_start(ap,cmd);
  arg=va_arg(ap,void *);
  va_end(ap);


  if (NULL == _socket_funcs.fcntl) {
    return -1;
  } /* if */

  sdp_sock_info = __sdp_sock_get(fd);

  __sdp_log(1, "FCNTL:      <%s:%d:%d> command <%d> argument <%d>\n",
            program_invocation_short_name, fd, __sdp_id_map(fd), cmd, 0);

  ret = _socket_funcs.fcntl(fd, cmd, arg);
  if (!(0 > ret) &&
      NULL != sdp_sock_info) {

    sret = _socket_funcs.fcntl(sdp_sock_info->fd, cmd, arg);
    if (sret < 0) {

      __sdp_log(9, "Error %d calling fcntl(%d, %d, %x) for SDP socket\n",
		sdp_sock_info->fd, cmd, 0, errno);
      __sdp_err_cleanup(fd);
    } /* if */
  } /* if */

  __sdp_log(1, "FCNTL:      <%s:%d:%d> result <%d>\n",
            program_invocation_short_name, fd, __sdp_id_map(fd), ret);

  return ret;
} /* fcntl */



/* ========================================================================= */
/*..socket -- replacement socket call. */
int socket(
           int domain,
           int type,
           int protocol
           )
{
#ifdef _SDP_VERBOSE_PRELOAD
  FILE *fd;
#endif
  struct sdp_socket_info *sdp_sock_info;
  int s;
  int t;


  if (simple_sdp_library) {
    if ((AF_INET == domain || AF_INET6 == domain) && SOCK_STREAM == type) {
      return _socket_funcs.socket(AF_INET_SDP, type, protocol);
    } else {
      return _socket_funcs.socket(domain, type, protocol);
    }
  }

  if (NULL == _socket_funcs.socket) {
    return -1;
  } /* if */

  __sdp_log(2, "SOCKET:     <%s> domain <%d> type <%d> protocol <%d>\n",
            program_invocation_short_name, domain, type, protocol);

  s = _socket_funcs.socket(domain, type, protocol);
  if (!(0 > s) &&
      MAPPED_SOCKET_MAX > s &&      0 == __sdp_config_empty()) {

    __sdp_struct_list[s] = NULL;

    if (((AF_INET == domain) || (AF_INET6 == domain)   ) && SOCK_STREAM == type) {

      t = _socket_funcs.socket(AF_INET_SDP, type, protocol);
      if (t < 0) {

	__sdp_log(9, "Error %d calling socket for SDP socket\n", errno);
      } /* if */
      else {

	sdp_sock_info = malloc(sizeof(struct sdp_socket_info));
	memset(sdp_sock_info, 0, sizeof(struct sdp_socket_info));

	sdp_sock_info->fd       = t;
	sdp_sock_info->refcount = 1;
	sdp_sock_info->state    = 0;
	sdp_sock_info->addr     = NULL;
	sdp_sock_info->turn     = 1;
	__sdp_struct_list[s] = sdp_sock_info;
      } /* else */
    } /* if */
  } /* if */

  __sdp_log(2, "SOCKET:     <%s:%d:%d>\n",
            program_invocation_short_name, s, __sdp_id_map(s));

  return s;
} /* socket */

/* ========================================================================= */
/*..setsockopt -- replacement setsockopt call. */
int setsockopt(
               int fd,
               int level,
               int optname,
               const void *optval,
               socklen_t optlen
               )
{
  struct sdp_socket_info *sdp_sock_info;
  int sret;
  int ret = 0;


  if (NULL == _socket_funcs.setsockopt) {
    return -1;
  } /* if */

  sdp_sock_info = __sdp_sock_get(fd);

  __sdp_log(1, "SETSOCKOPT: <%s:%d:%d> level <%d> name <%d>\n",
            program_invocation_short_name, fd, __sdp_id_map(fd),
	    level, optname);

  ret = _socket_funcs.setsockopt(fd, level, optname, optval, optlen);
  if (!(0 > ret) &&
      NULL != sdp_sock_info) {

    sret = _socket_funcs.setsockopt(sdp_sock_info->fd, level,
				    optname, optval, optlen);
    if (sret < 0) {
      
      __sdp_log(9, "Error %d calling setsockopt for SDP socket, closing\n",
		errno);
      __sdp_err_cleanup(fd);
    } /* if */
  } /* if */

  __sdp_log(1, "SETSOCKOPT: <%s:%d:%d> result <%d>\n",
            program_invocation_short_name, fd, __sdp_id_map(fd), ret);

  return ret;
} /* setsockopt */

/* ========================================================================= */
/*..connect -- replacement connect call. */
int connect(
            int fd,
            const struct sockaddr *serv_addr,
            socklen_t addrlen
            )
{
  const struct sockaddr_in *sin = (const struct sockaddr_in *)serv_addr;
  int replaced = 0;
  const struct sockaddr_in6 *sock_ipv6 = (const struct sockaddr_in6 *)serv_addr;
  struct sockaddr_in sockaddr_conv;
  u_int16_t  ipv4_port = sock_ipv6->sin6_port;
  u_int32_t ipv4address;
  unsigned char *ipv6address;
  int ret;


  memcpy(&sockaddr_conv, serv_addr, sizeof sockaddr_conv);

  ipv6address = (unsigned char *)&(sock_ipv6->sin6_addr.s6_addr);
  memcpy(&ipv4address,&ipv6address[12],4);



  __sdp_log(1, "CONNECT:    <%s:%d:%d> type <%d> IP <%d.%d.%d.%d> port <%d>\n",
            program_invocation_short_name, fd, __sdp_id_map(fd),
	    sin->sin_family,
	    (AF_INET == sin->sin_family ?
	     (sin->sin_addr.s_addr) & 0xff : 0),
	    (AF_INET == sin->sin_family ?
	     (sin->sin_addr.s_addr >>  8) & 0xff : 0),
	    (AF_INET == sin->sin_family ?
	     (sin->sin_addr.s_addr >> 16) & 0xff : 0),
	    (AF_INET == sin->sin_family ?
	     (sin->sin_addr.s_addr >> 24) & 0xff : 0),
    	    (AF_INET == sin->sin_family ? ntohs(sin->sin_port) : 0));

  if (NULL == _socket_funcs.connect) {
    return -1;
  } /* if */
  if (simple_sdp_library) {
     return  _socket_funcs.connect(fd, serv_addr, addrlen);
  }
  
  if (!(0 > fd) &&
      MAPPED_SOCKET_MAX > fd) {

    replaced=decide_family(fd, 1, sin);
  } /* if */

  if (sin->sin_family==AF_INET6 && replaced==1 ) {
    struct in_addr new_address;
   
    sockaddr_conv.sin_family=AF_INET;
    sockaddr_conv.sin_port=ipv4_port;

    new_address.s_addr=ipv4address;

    sockaddr_conv.sin_addr= new_address;
    ret=_socket_funcs.connect(fd, (struct sockaddr *)&sockaddr_conv, 16);
  }
  else
    ret = _socket_funcs.connect(fd, serv_addr, addrlen);

  __sdp_log(1, "CONNECT:    <%s:%d:%d> result <%d>\n",
            program_invocation_short_name, fd, __sdp_id_map(fd), ret);

  return ret;
} /* connect */

/* ========================================================================= */
/*..listen -- replacement listen call. */
int listen(
           int fd,
           int backlog
           )
{
  struct sdp_socket_info *sdp_sock_info;
  int ret;
  int both_listen=0;


  if (NULL == _socket_funcs.listen) {
    return -1;
  } /* if */


  if (simple_sdp_library) {
    return _socket_funcs.listen(fd, backlog);
  }
  sdp_sock_info = __sdp_sock_get(fd);
  __sdp_log(1, "LISTEN:     <%s:%d:%d>\n",
            program_invocation_short_name, fd, __sdp_id_map(fd));

  if (sdp_sock_info!=NULL) {
    both_listen=__sdp_match_listen(sdp_sock_info->addr,1);
  };

  if ( both_listen==0) {
    if (NULL != sdp_sock_info &&
	0 < (STATE_BIND_CALLED & sdp_sock_info->state)) {
      
      decide_family(fd, 0, sdp_sock_info->addr);
    } /* if */
    
    ret = _socket_funcs.listen(fd, backlog);
  } else {

    ret = _socket_funcs.listen(fd, backlog);
    
    if ((ret >= 0) && sdp_sock_info)
      {
	ret = _socket_funcs.listen(sdp_sock_info->fd, backlog);  
      }
  }
  
  
  __sdp_log(1, "LISTEN:     <%s:%d:%d> result <%d>\n",
            program_invocation_short_name, fd, __sdp_id_map(fd), ret);

  return ret;
} /* listen */

/* setsockopt() level and optname declarations */
#define SOL_SDP		1025
#define SDP_UNBIND	259	/* Unbind socket */

/* ========================================================================= */
/*..bind -- replacement bind call. */
int bind(
         int fd,
         const struct sockaddr *my_addr,
         socklen_t addrlen
         )
{
  const struct sockaddr_in *sin = (const struct sockaddr_in *)my_addr;
  struct sdp_socket_info *sdp_sock_info;
  int sret;

  if (NULL == _socket_funcs.bind) {
    return -1;
  } /* if */

  if (simple_sdp_library) {
    return _socket_funcs.bind(fd, my_addr, addrlen);
  }

  sdp_sock_info = __sdp_sock_get(fd);

  __sdp_log(1, "BIND:       <%s:%d:%d> type <%d> IP <%d.%d.%d.%d> port <%d>\n",
            program_invocation_short_name, fd, __sdp_id_map(fd),
	    sin->sin_family,
	    (AF_INET == sin->sin_family ?
	     (sin->sin_addr.s_addr) & 0xff : 0),
	    (AF_INET == sin->sin_family ?
	     (sin->sin_addr.s_addr >>  8) & 0xff : 0),
	    (AF_INET == sin->sin_family ?
	     (sin->sin_addr.s_addr >> 16) & 0xff : 0),
	    (AF_INET == sin->sin_family ?
	     (sin->sin_addr.s_addr >> 24) & 0xff : 0),
    	    (AF_INET == sin->sin_family ? ntohs(sin->sin_port) : 0));


  
  if (NULL == sdp_sock_info) {

    sret = _socket_funcs.bind(fd, my_addr, addrlen);
  } /* if */
  else {
    /*
     * Valid SDP socket.
     */
    if (0 < (STATE_LISTEN_CALLED & sdp_sock_info->state)) {

      decide_family(fd, 0, sin);
    } /* if */

    sdp_sock_info->state |= STATE_BIND_CALLED;

    if (ntohs(sin->sin_port) == 0) {
      struct sockaddr_in tsin;
      socklen_t tsinlen = sizeof(tsin);
      unsigned int tries;
      /*
       * The application is requesting a port be assigned for the connection.
       * To make sure everything works, we require both the SDP and TCP ports
       * get allocated the same port number.
       */
      for (tries = 0; tries < MAX_BIND_TRIES; tries++) {
	/* Bind to TCP socket first */
	sret = _socket_funcs.bind(fd, my_addr, addrlen);
	if (sret < 0) {
	  /* We can't really do anything with the SDP socket in this case */
	  goto done;
	} /* if */
	
	/* Get the allocated port number */
	if (getsockname(fd, (struct sockaddr *)&tsin, &tsinlen) < 0) {
	  __sdp_log(9, "Error <%d> getsockname for TCP socket, closing SDP\n",
		  errno);
	  __sdp_err_cleanup(fd);

	  sret = 0;
	  goto done;
	} /* if */

	/* Bind to SDP socket last */
	sret = _socket_funcs.bind(sdp_sock_info->fd,
				  (struct sockaddr *)&tsin,
				  tsinlen);
	if (!sret) {
	  /* Success */
	  break;
	} /* if */
      } /* for */

      if (tries >= MAX_BIND_TRIES) {
	/* We gave up */
	errno = EADDRINUSE;
	sret = -1;
	goto done;
      } /* if */

      sin = &tsin;
    } /* if */
    else {
      /*
       * Apply operation to both sockets. We want both to succeed, or we return
       * an error. Falling back to TCP is causing problems since remote systems
       * may not be falling back as well.
       */
      sret = _socket_funcs.bind(sdp_sock_info->fd, my_addr, addrlen);
      if (sret < 0) {

	__sdp_log(9, "Error %d calling bind() for SDP socket\n", errno);
	goto done;
      } /* if */
  
      sret = _socket_funcs.bind(fd, my_addr, addrlen);
      if (sret < 0) {
	__sdp_log(9, "Error %d calling bind() for TCP socket\n", errno);
	_socket_funcs.setsockopt(sdp_sock_info->fd,
				 SOL_SDP, SDP_UNBIND, NULL, 0);
	goto done;
      } /* if */
    } /* else */

    sdp_sock_info->addr = malloc(sizeof(struct sockaddr_in));
    if (NULL != sdp_sock_info->addr) {
      
      memcpy(sdp_sock_info->addr, sin, sizeof(struct sockaddr_in));
    } /* if */

    sret = 0;
  } /* else */

done:
  __sdp_log(1, "BIND:       <%s:%d:%d> result <%d>\n",
            program_invocation_short_name, fd, __sdp_id_map(fd), sret);

  return sret;
} /* bind */

/* ========================================================================= */
/*..close -- replacement close call. */
int close(
          int fd
          )
{
  struct sdp_socket_info *sdp_sock_info;


  if (NULL == _socket_funcs.close) {
    return -1;
  } /* if */

  sdp_sock_info = __sdp_sock_get(fd);

  __sdp_log(1, "CLOSE:      <%s:%d:%d>\n",
            program_invocation_short_name, fd, __sdp_id_map(fd));

  if (NULL != sdp_sock_info) {

    sdp_sock_info->refcount -= 1;

    if (0 == sdp_sock_info->refcount) {

      _socket_funcs.close(sdp_sock_info->fd);

      if (NULL != sdp_sock_info->addr) {

	free(sdp_sock_info->addr);
      } /* if */

      free(sdp_sock_info);
    } /* if */

    __sdp_struct_list[fd] = NULL;
  } /* if */

  return _socket_funcs.close(fd);
} /* close */

/* ========================================================================= */
/*..dup -- replacement socket call. */
int dup(
	int fd
	)
{
  struct sdp_socket_info *sdp_sock_info;
  int newfd;


  if (NULL == _socket_funcs.socket) {
    return -1;
  } /* if */

  sdp_sock_info = __sdp_sock_get(fd);

  __sdp_log(1, "DUP:        <%s:%d:%d>\n",
            program_invocation_short_name, fd, __sdp_id_map(fd));

  newfd = _socket_funcs.dup(fd);
  if (!(0 > newfd) &&
      NULL != sdp_sock_info) {

    if (MAPPED_SOCKET_MAX > newfd) {

      sdp_sock_info->refcount  += 1;
      __sdp_struct_list[newfd]  = sdp_sock_info;
    } /* if */
    else {

      __sdp_log(9, "DUP: new fd <%d> out of range.\n", newfd);
      __sdp_err_cleanup(fd);
    } /* else */
  } /* if */

  __sdp_log(1, "DUP:        <%s:%d:%d:%d>\n",
            program_invocation_short_name, fd, newfd, __sdp_id_map(fd));

  return newfd;
} /* dup */

/* ========================================================================= */
/*..getsockname -- replacement getsocknanme call. */
int getsockname(int fd,
		struct sockaddr *name,
		socklen_t *namelen)
{
  int ret = 0;

  /*
   * ensure the SDP protocol family is not exposed to the user, since
   * this is meant to be a transparency layer.
   */
  if (NULL == _socket_funcs.getsockname) {
    return -1;
  } /* if */

  __sdp_log(1, "GETSOCKNAME <%s:%d>\n",
	    program_invocation_short_name, fd);

  ret = _socket_funcs.getsockname(fd, name, namelen);
  if (!(0 > ret)) {
    /*
     * parameters will not be null/invalid if we have a good
     * return code from the function
     */
    if (AF_INET_SDP == name->sa_family) {

      name->sa_family = AF_INET;
    } /* if */
  } /* if */

  __sdp_log(1, "GETSOCKNAME <%s:%d> result <%d:%d>\n",
	    program_invocation_short_name, fd, ret,
	    (!(0 > ret) ?  : -1));

  return ret;
} /* getsockname */

/* ========================================================================= */
/*..getpeername -- replacement getpeername call. */
int getpeername(int fd,
		struct sockaddr *name,
		socklen_t *namelen)
{
  int ret = 0;

  /*
   * ensure the SDP protocol family is not exposed to the user, since
   * this is meant to be a transparency layer.
   */

  if (NULL == _socket_funcs.getpeername) {
    return -1;
  } /* if */

  

  __sdp_log(1, "GETPEERNAME <%s:%d>\n",
	    program_invocation_short_name, fd);

  ret = _socket_funcs.getpeername(fd, name, namelen);
  if (!(0 > ret)) {
    /*
     * parameters will not be null/invalid if we have a good
     * return code from the function
     */
    if (AF_INET_SDP == name->sa_family) {

      name->sa_family = AF_INET;
    } /* if */
  } /* if */

  __sdp_log(1, "GETPEERNAME <%s:%d> result <%d:%d> family=%d s_addr=%d\n",
	    program_invocation_short_name, fd, ret,
	    (!(0 > ret) ?  : -1),name->sa_family,
            ((struct sockaddr_in*)name)->sin_addr.s_addr);

  return ret;
} /* getpeername */



/* ========================================================================= */
/*..accept -- replacement accept call. */
int accept(int fd,
		struct sockaddr *addr,
		socklen_t *addrlen)
{
  int ret = 0;
  fd_set fds;
  struct sdp_socket_info *sdp_sock_info;

  /*
   * ensure the SDP protocol family is not exposed to the user, since
   * this is meant to be a transparency layer.
   */
  if (NULL == _socket_funcs.accept) {
    return -1;
  } /* if */

  __sdp_log(1, "ACCEPT <%s:%d>\n",
	    program_invocation_short_name, fd);

  sdp_sock_info = __sdp_sock_get(fd);

  if (sdp_sock_info == NULL) {
    ret = _socket_funcs.accept(fd, addr, addrlen);
  }
  else {
    FD_ZERO(&fds);
    FD_SET(fd,&fds);
    FD_SET(sdp_sock_info->fd,&fds);
    ret = _socket_funcs.select( 1+ ( (fd>sdp_sock_info->fd) ? fd : sdp_sock_info->fd ) , 
				&fds, NULL, NULL, NULL);
    
    if (ret >= 0) { 
      if (sdp_sock_info->turn==1) {
	sdp_sock_info->turn=0;
	if      (FD_ISSET(fd,&fds)) {
	  ret = _socket_funcs.accept(fd, addr, addrlen);
	}
	else {
	  ret = _socket_funcs.accept(sdp_sock_info->fd, addr, addrlen);
	}
      } else {
	sdp_sock_info->turn=1;
	if (FD_ISSET(sdp_sock_info->fd,&fds)) {
	  ret = _socket_funcs.accept(sdp_sock_info->fd, addr, addrlen);
	}  else  {
	  ret = _socket_funcs.accept(fd, addr, addrlen);
	};
      }
    }
      else {
	__sdp_log(9, "Error select finished but no FD is ready\n");
      }
  }
 
  
  if (!(0 > ret)) {
    /*
     * parameters will not be null/invalid if we have a good
     * return code from the function
     */
    if (addr != NULL && AF_INET_SDP == addr->sa_family) { 
	 
         addr->sa_family = AF_INET;
       } /* if */
  } /* if */
  return ret;
} /* accept */




/* ========================================================================= */
/*..select -- replacement socket call. */
int select(
       int n,  fd_set  *readfds,  fd_set  *writefds,
       fd_set *exceptfds, struct timeval *timeout
	)
{
  int ret;
  int current;
  int maxi=0;
  fd_set new_fds;
  FD_ZERO(&new_fds);
  struct sdp_socket_info *sdp_sock_info;
  if (readfds == NULL) {
    return _socket_funcs.select(n,readfds, writefds, exceptfds , timeout);
  }
  if (n>0) {
    maxi=n-1;
  }
  
 for (current = 0; current<n; current++) {
    sdp_sock_info = __sdp_sock_get(current);
    if (FD_ISSET(current,readfds)) {
      FD_SET(current,&new_fds);
      if (current>maxi) {
	maxi=current;
      };
      if (sdp_sock_info != NULL) {
	FD_SET(sdp_sock_info->fd,&new_fds);
	if (sdp_sock_info->fd > maxi) {
	  maxi=sdp_sock_info->fd;
	};
      };
    };
 };
      
 ret =_socket_funcs.select( 1+ maxi , 
	       &new_fds, writefds, exceptfds , timeout);
 
 if (ret<0) {
   return ret;
 };
 for (current = 0; current<n; current++) {
   sdp_sock_info = __sdp_sock_get(current);
   if (sdp_sock_info==NULL) {
     if (FD_ISSET(current,readfds) && FD_ISSET(current,&new_fds)==0) {
       FD_CLR(current,readfds);
     };
   } else {
     if (FD_ISSET(current,readfds) &&   FD_ISSET(current,&new_fds) &&  FD_ISSET(sdp_sock_info->fd,&new_fds)) {
       ret-=1;
     };
     if (FD_ISSET(current,readfds) &&  FD_ISSET(current,&new_fds)==0 &&   FD_ISSET(sdp_sock_info->fd,&new_fds)==0 ) {
       FD_CLR(current,readfds);
     };
   };
 };
 
 return ret;
 
} /* select */

/* ========================================================================= */







/* --------------------------------------------------------------------- */
/*                                                                       */
/* Library load/unload initialization/cleanup                            */
/*                                                                       */
/* --------------------------------------------------------------------- */
/* ========================================================================= */
/*..__sdp_init -- intialize the library */
void __sdp_init(
                void
                )
{
  char *config_file, *error_str;
  int counter;


  for (counter = 0; counter < MAPPED_SOCKET_MAX; counter++) {

    __sdp_struct_list[counter] = NULL;
  } /* for */
  /*
   * FIXME: Avoid using fprintf unless explicitly configured by the user.
   * stderr (fd 2) may have been closed by someone else already.
   */



#ifndef RTLD_NEXT
  /*
   * open libc for original socket call.
   */
  __libc_dl_handle = dlopen("/lib64/libc.so.6", RTLD_LAZY);
  if (NULL == __libc_dl_handle) {
    __libc_dl_handle = dlopen("/lib/libc.so.6", RTLD_LAZY);
    if (NULL == __libc_dl_handle) {
      fprintf(stderr, "%s\n", dlerror());
      exit(-1);
    }
  } /* if */
#endif



  /*
   * Get the original functions
   */
  _socket_funcs.ioctl = dlsym(__libc_dl_handle, "ioctl");
  if (NULL != (error_str = dlerror())) {
    fprintf(stderr, "%s\n", error_str);
    exit(-1);
  } /* if */

  _socket_funcs.fcntl = dlsym(__libc_dl_handle, "fcntl");
  if (NULL != (error_str = dlerror())) {
    fprintf(stderr, "%s\n", error_str);
    exit(-1);
  } /* if */

  _socket_funcs.socket = dlsym(__libc_dl_handle, "socket");
  if (NULL != (error_str = dlerror())) {
    fprintf(stderr, "%s\n", error_str);
    exit(-1);
  } /* if */

  _socket_funcs.setsockopt = dlsym(__libc_dl_handle, "setsockopt");
  if (NULL != (error_str = dlerror())) {
    fprintf(stderr, "%s\n", error_str);
    exit(-1);
  } /* if */

  _socket_funcs.connect = dlsym(__libc_dl_handle, "connect");
  if (NULL != (error_str = dlerror())) {
    fprintf(stderr, "%s\n", error_str);
    exit(-1);
  } /* if */

  _socket_funcs.listen = dlsym(__libc_dl_handle, "listen");
  if (NULL != (error_str = dlerror())) {
    fprintf(stderr, "%s\n", error_str);
    exit(-1);
  } /* if */

  _socket_funcs.bind = dlsym(__libc_dl_handle, "bind");
  if (NULL != (error_str = dlerror())) {
    fprintf(stderr, "%s\n", error_str);
    exit(-1);
  } /* if */

  _socket_funcs.close = dlsym(__libc_dl_handle, "close");
  if (NULL != (error_str = dlerror())) {
    fprintf(stderr, "%s\n", error_str);
    exit(-1);
  } /* if */

  _socket_funcs.dup = dlsym(__libc_dl_handle, "dup");
  if (NULL != (error_str = dlerror())) {
    fprintf(stderr, "%s\n", error_str);
    exit(-1);
  } /* if */

  _socket_funcs.getpeername = dlsym(__libc_dl_handle, "getpeername");
  if (NULL != (error_str = dlerror())) {
    fprintf(stderr, "%s\n", error_str);
    exit(-1);
  } /* if */

  _socket_funcs.getsockname = dlsym(__libc_dl_handle, "getsockname");
  if (NULL != (error_str = dlerror())) {
    fprintf(stderr, "%s\n", error_str);
    exit(-1);
  } /* if */

  _socket_funcs.accept = dlsym(__libc_dl_handle, "accept");
  if (NULL != (error_str = dlerror())) {
    fprintf(stderr, "%s\n", error_str);
    exit(-1);
  } /* if */

  _socket_funcs.select = dlsym(__libc_dl_handle, "select");
  if (NULL != (error_str = dlerror())) {
    fprintf(stderr, "%s\n", error_str);
    exit(-1);
  } /* if */
  if (getenv("SIMPLE_LIBSDP")==NULL) {
    simple_sdp_library=0;
  } else {
    simple_sdp_library=1;
    return;
  };
  config_file = getenv("LIBSDP_CONFIG_FILE");
  if (config_file) {
    __sdp_read_config(config_file);
  } else {
      printf("default libsdp configuration is used\n");
#define LIBSDP_DEFAULT_CONFIG_FILE  "/usr/local/ibgd/etc/libsdp.conf"
    __sdp_read_config(LIBSDP_DEFAULT_CONFIG_FILE);
  }
} /* __sdp_init */

/* ========================================================================= */
/*..__sdp_fini -- when the library is unloaded this is called */
void __sdp_fini(
              void
              )
{
#ifndef RTLD_NEXT
  dlclose(__libc_dl_handle);
#endif
} /* _fini */
